/*
 * This module is used to trigger a segment not-present exception when handling
 * an interrupt through the IDT. There are two experiments that can be done
 * with this module.
 *
 * 0. Baseline
 * sysctl debug.intrfault.present=1	# no nested exception while handling
 *					# interrupts
 * sysctl debug.intrfault.idtvec	# verify 'idtvec' is non-zero
 * sysctl debug.intrfault.count		# baseline count of 'idtvec' interrupts
 * sysctl debug.intrfault.swintr=1	# verify debug.intrfault.count bumped
 * sysctl debug.intrfault.ipi=<cpu>	# verify debug.intrfault.count bumped
 *
 * 1. Software interrupt handling with nested exception (#NP)
 * sysctl debug.intrfault.present=0	# set P bit to 0 in idtvec gate
 * sysctl debug.intrfault.swintr=1	# trigger a s/w interrupt at idtvec
 *
 * This will enter the debugger and hitting 'c' in DDB will keep hitting
 * the same exception over and over again. This is because the %rip saved
 * in the #NP trapframe points to the "int $idtvec" instruction. There is
 * no loss of information in this case.
 *
 * 2. Hardware interrupt handling with nested exception (#NP)
 * sysctl debug.intrfault.present=0	# set P bit to 0 in idtvec gate
 * sysctl debug.intrfault.ipi=<cpu>	# trigger ipi to <cpu> at vector idtvec
 *
 * This will enter the debugger and hitting 'c' in DDB will silently continue.
 * In other words the fact that a hardware interrupt was being delivered to
 * 'idtvec' is forgotten after the #NP exception handler returns.
 */

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/lock.h>
#include <sys/module.h>
#include <sys/mutex.h>
#include <sys/sysctl.h>
#include <sys/systm.h>
#include <sys/smp.h>
#include <sys/proc.h>
#include <sys/pcpu.h>
#include <sys/sched.h>

#include <machine/intr_machdep.h>
#include <machine/segments.h>
#include <machine/md_var.h>
#include <machine/pcpu.h>
#include <machine/smp.h>
#include <machine/apicvar.h>

extern inthand_t IDTVEC(rsvd);
extern inthand_t IDTVEC(intrfault_swint_handler);
extern inthand_t IDTVEC(intrfault_ipi_handler);
extern void intrfault_handler(void);

static SYSCTL_NODE(_debug, OID_AUTO, intrfault, CTLFLAG_RW, 0, "");

static int idtvec;
SYSCTL_INT(_debug_intrfault, OID_AUTO, idtvec, CTLFLAG_RD, &idtvec, 0, NULL);

static int count;
SYSCTL_INT(_debug_intrfault, OID_AUTO, count, CTLFLAG_RW, &count, 0, NULL);

static int present;
SYSCTL_INT(_debug_intrfault, OID_AUTO, present, CTLFLAG_RW, &present, 0, NULL);

void
intrfault_handler(void)
{
	count++;
}

CTASSERT(APIC_IPI_INTS == 243);
CTASSERT(APIC_SPURIOUS_INT == 255);

static void
swintr(int vector)
{
	KASSERT(vector >= APIC_IPI_INTS && vector < APIC_SPURIOUS_INT,
	    ("invalid vector %d", vector));

	switch (vector) {
	case 243:
		__asm __volatile("int $243");
		break;
	case 244:
		__asm __volatile("int $244");
		break;
	case 245:
		__asm __volatile("int $245");
		break;
	case 246:
		__asm __volatile("int $246");
		break;
	case 247:
		__asm __volatile("int $247");
		break;
	case 248:
		__asm __volatile("int $248");
		break;
	case 249:
		__asm __volatile("int $249");
		break;
	case 250:
		__asm __volatile("int $250");
		break;
	case 251:
		__asm __volatile("int $251");
		break;
	case 252:
		__asm __volatile("int $252");
		break;
	case 253:
		__asm __volatile("int $253");
		break;
	case 254:
		__asm __volatile("int $254");
		break;
	default:
		panic("swintr: invalid vector %d", vector);
	}
}

static int
sysctl_swintr(SYSCTL_HANDLER_ARGS)
{
	struct gate_descriptor *ip;
	int error, val;

	/* If a vector could not be allocated then just return with an error */
	if (idtvec == 0)
		return (EBUSY);

	val = 0;
	error = sysctl_handle_int(oidp, &val, 0, req);
	if (error || req->newptr == NULL || val == 0)
		return (error);

#ifdef __i386__
	setidt(idtvec, IDTVEC(intrfault_swint_handler), SDT_SYS386IGT, SEL_KPL,
	    GSEL(GCODE_SEL, SEL_KPL));
#else
	setidt(idtvec, IDTVEC(intrfault_swint_handler), SDT_SYSIGT, SEL_KPL, 0);
#endif
	ip = &idt[idtvec];
	ip->gd_p = present ? 1 : 0;
	mb();

	swintr(idtvec);
	return (0);
}

SYSCTL_PROC(_debug_intrfault, OID_AUTO, swintr, CTLTYPE_INT|CTLFLAG_RW, 0, 0, 
    sysctl_swintr, "I", "");

static int
sysctl_ipi(SYSCTL_HANDLER_ARGS)
{
	struct gate_descriptor *ip;
	int error, cpu;

	/* If a vector could not be allocated then just return with an error */
	if (idtvec == 0)
		return (EBUSY);

	cpu = 0;
	error = sysctl_handle_int(oidp, &cpu, 0, req);
	if (error || req->newptr == NULL)
		return (error);

	if (cpu < 0 || cpu >= mp_ncpus)
		return (EINVAL);

#ifdef __i386__
	setidt(idtvec, IDTVEC(intrfault_ipi_handler), SDT_SYS386IGT, SEL_KPL,
	    GSEL(GCODE_SEL, SEL_KPL));
#else
	setidt(idtvec, IDTVEC(intrfault_ipi_handler), SDT_SYSIGT, SEL_KPL, 0);
#endif
	ip = &idt[idtvec];
	ip->gd_p = present ? 1 : 0;
	mb();

	if (!present) {
		/* 
		 * If the P bit in the interrupt gate is 0 then it is expected
		 * we'll hit the nested #NP exception instead of executing the
		 * interrupt handler code which signals an EOI to the lapic.
		 * The EOI must be signalled otherwise low priority interrupts
		 * will be blocked permanently.
		 *
		 * To ensure that we signal an EOI on the correct CPU we bind
		 * this thread to 'cpu' and also enter a critical section to
		 * guarantee that this thread will not be preempted.
		 */
		thread_lock(curthread);
		sched_bind(curthread, cpu);
		thread_unlock(curthread);
		critical_enter();
	}

	ipi_cpu(cpu, idtvec);

	if (!present) {
		/*
		 * Wait until the IPI is delivered before signalling an EOI
		 * to the APIC. Otherwise this creates a race where the EOI
		 * might happen before the APIC moves 'idtvec' into the ISR
		 * causing a subsequent hang.
		 */
		lapic_ipi_wait(-1);

		lapic_eoi();

		critical_exit();
		thread_lock(curthread);
		sched_unbind(curthread);
		thread_unlock(curthread);
	}
	return (0);
}

SYSCTL_PROC(_debug_intrfault, OID_AUTO, ipi, CTLTYPE_INT|CTLFLAG_RW, 0, 0, 
    sysctl_ipi, "I", "");

static int
idtvec_alloc(void)
{
	int idx;
	uintptr_t func;
	struct gate_descriptor *ip;

	/*
	 * Search backwards from the highest IDT vector available for use
	 * as our vector. We install the 'intrfault' handler at that vector
	 * and use it to interrupt the vcpus.
	 */
	idx = APIC_SPURIOUS_INT;
	while (--idx >= APIC_IPI_INTS) {
		ip = &idt[idx];
		func = ((long)ip->gd_hioffset << 16 | ip->gd_looffset);
		if (func == (uintptr_t)&IDTVEC(rsvd))
			return (idx);
	}
	return (0);
}

static void
idtvec_free(int vector)
{
	if (vector == 0)
		return;

	KASSERT(vector >= APIC_IPI_INTS && vector < APIC_SPURIOUS_INT,
	    ("invalid vector %d", vector));

	setidt(vector, IDTVEC(rsvd), SDT_SYSIGT, SEL_KPL, 0);
}

static int
intrfault_modevent(module_t mod, int type, void *unused)
{

        switch (type) {
        case MOD_LOAD:
		idtvec = idtvec_alloc();
		break;
	case MOD_UNLOAD:
		idtvec_free(idtvec);
		break;
	default:
		return (EINVAL);
	}

	return (0);
}

static moduledata_t intrfault_mod_data= {
	"intrfault",
	intrfault_modevent,
};

DECLARE_MODULE(test, intrfault_mod_data, SI_SUB_SMP + 1, SI_ORDER_ANY);
